import logging
from functools import partial
from typing import (
    Generator,
    List,
    Tuple,
    Union,
)

from facebook_business.api import (
    FacebookAdsApi,
    FacebookRequest,
    FacebookResponse,
)
from facebook_business.exceptions import FacebookRequestError

from wespe.exceptions import (
    BatchExecutionError,
    InvalidValueError,
    NoFacebookRequestProvidedError,
    TooManyRequestsPerBatchError,
)
from .requests import (
    BaseRequestError,
    BaseResponse,
)

logger = logging.getLogger(__name__)


class FacebookBatchResponse(BaseResponse):
    def __init__(self, request: FacebookRequest, response: FacebookResponse):
        super().__init__(
            data=response.json()
        )

        self.request = request
        self.response = response

    def __str__(self):
        return 'FacebookResponse: {}. When using params: {}.'.format(self.data, self.request.get_params())


class FacebookBatchRequestError(BaseRequestError):
    def __init__(self, request: FacebookRequest, request_error: FacebookRequestError):
        super().__init__(
            description=request_error.api_error_message(),
            is_transient=request_error.api_transient_error(),
            data=request_error.error().body()['error']
        )

        self.request = request
        self.request_error = request_error

    def __str__(self):
        return 'FacebookRequestError: {}. When using params: {}.'.format(self.data, self.request.get_params())


class FacebookBatch:
    def __init__(self, requests: List[FacebookRequest], api: FacebookAdsApi):
        if not requests:
            raise NoFacebookRequestProvidedError('At least one facebook request must be provided')

        num_requests = len(requests)

        if num_requests >= 50:
            # See more on https://developers.facebook.com/docs/graph-api/making-multiple-requests
            raise TooManyRequestsPerBatchError('A maximum of 50 requests per batch is supported')

        self._batch = api.new_batch()
        self._requests = requests
        self._responses = [None] * num_requests
        self._errors = [None] * num_requests

    def execute(self):
        """
        Execute all requests.

        :return:
        """
        for request_index, request in enumerate(self._requests):
            self._batch.add_request(
                request,
                success=partial(self._default_success_callback, request_index=request_index),
                failure=partial(self._default_failure_callback, request_index=request_index))

        self._batch.execute()

    def _default_failure_callback(
        self, response: FacebookResponse, request_index: int, object_id: int=None
    ):
        """
        A method that can be used to raise exceptions when the batch object is used for bulk operations.
        This callback raises a custom FacebookAPIError.

        This is intended to be used as a default callback to fallback to in case a user does not provide one.

        :param response: Facebook response object.
        :param request_index: The index of the request in the whole batch.
        :param object_id: (Optional) The ID of the object being updated.
        """
        request = self._requests[request_index]
        batch__error = FacebookBatchRequestError(request=request, request_error=response.error())
        self._errors[request_index] = batch__error

        error_msg = ['#{} -'.format(request_index)]

        if object_id:
            error_msg.append('Error updating object with id [{}].'.format(object_id))

        error_msg.append(str(batch__error))

        logger.error(' '.join(error_msg))

    def _default_success_callback(
        self, response: FacebookResponse, request_index: int, object_id: int=None
    ):
        """
        A method that can be used to log when the batch object has completed successfully.

        This is intended to be used as a default callback to fallback to in case a user does not provide one.

        :param response: Facebook response object.
        :param request_index: The index of the request in the whole batch.
        :param object_id: The ID of the object being updated.
        """
        request = self._requests[request_index]
        batch_response = FacebookBatchResponse(request=request, response=response)
        self._responses[request_index] = batch_response

        if object_id is None:
            object_id = response.json().get('id')

        logger.debug('Request #{}: Object with id [{}] updated successfully!'.format(
            request_index, str(object_id)
        ))


class FacebookBatchUploader:
    def __init__(self, requests: List[FacebookRequest], api: FacebookAdsApi=None):
        self.api = api or FacebookAdsApi.get_default_api()

        self._requests = requests
        self._batches = []

    def execute(self, chunk_size: int=50):
        """
        Execute all requests in batches of chunk_size amount.

        :param chunk_size:
            (Optional) The amount of requests per chunk. Keep in mind this value should be between 1 and 50, otherwise
            an exception will be raised. Defaults to 50.

        :return:
        """
        num_requests = len(self.requests)

        if chunk_size < 1 or chunk_size > 50:
            raise InvalidValueError('Chunk size must be between 1 and 50')

        for i in range(0, num_requests, chunk_size):
            batch = FacebookBatch(self.requests[i:i + chunk_size], api=self.api)
            self._batches.append(batch)
            batch.execute()

        failed_requests = list(self.errors)

        if any(failed_requests):
            exception_msg = '{} requests failed out of {}'.format(len(failed_requests), num_requests)
            raise BatchExecutionError(exception_msg)

    @property
    def requests(self) -> List[FacebookRequest]:
        """
        Returns all FacebookRequest instances.

        :return: a list of FacebookRequest instances.
        """
        return self._requests

    @property
    def responses(self) -> Generator[Union[None, FacebookBatchResponse], None, None]:
        """
        Returns the responses for the executed FacebookRequest instances. The amount and order of elements will
        be the same as of FacebookRequest instances. FacebookRequest with errors will have None as their value
        in the respective index.

        :return: a list of FacebookBatchResponse instance and/or None.
        """
        for batch in self._batches:
            for response in batch._responses:
                yield response

    @property
    def errors(self) -> Generator[Union[None, FacebookBatchRequestError], None, None]:
        """
        Returns the errors for the executed FacebookRequest instances. The amount and order of elements will
        be the same as of FacebookRequest instances. FacebookRequest without errors will have None as their value
        in the respective index.

        :return: a generator of FacebookBatchRequestError instances and/or None.
        """
        for batch in self._batches:
            for response in batch._errors:
                yield response

    @property
    def items(self) -> Generator[
        Tuple[FacebookRequest, Union[None, FacebookBatchResponse], Union[None, FacebookBatchRequestError]], None, None
    ]:
        """
        Returns a list of tuples shaped as FacebookRequest instance and its respective response/error.

        :return: a list of tuples shaped as (request, FacebookBatchResponse or None, FacebookBatchRequestError
                 instance or None).
        """
        for request, response, error in zip(self.requests, self.responses, self.errors):
            yield request, response, error



# coding: utf8
from itertools import chain
from unittest import TestCase
from unittest.mock import (
    ANY,
    MagicMock,
    PropertyMock,
    call,
    patch,
)

from facebook_business import FacebookAdsApi

from wespe.batch_uploaders import FacebookBatchUploader
from wespe.batch_uploaders.facebook import (
    FacebookBatch,
    FacebookBatchRequestError,
    FacebookBatchResponse,
)
from wespe.exceptions import (
    BatchExecutionError,
    InvalidValueError,
    NoFacebookRequestProvidedError,
    TooManyRequestsPerBatchError,
)


class TestFacebookBatch(TestCase):
    def setUp(self):
        self.requests = [MagicMock()] * 10
        self.api = MagicMock()
        self.batch = FacebookBatch(requests=self.requests, api=self.api)

    def test_constructor_raises_NoFacebookRequestProvidedError_when_request_list_is_empty(self):
        with self.assertRaises(NoFacebookRequestProvidedError):
            FacebookBatch(requests=[], api=self.api)

    def test_constructor_raises_TooManyRequestsPerBatchError_when_more_than_50_requests_were_provided(self):
        with self.assertRaises(TooManyRequestsPerBatchError):
            FacebookBatch(requests=[MagicMock()] * 51, api=self.api)

    def test_constructor_initialiazes_a_batch_using_provided_api(self):
        self.assertEqual(self.batch._batch, self.api.new_batch.return_value)

    def test_constructor_initialiazes_a_list_of_None_for_errors_prior_to_execution(self):
        self.assertListEqual(self.batch._errors, [None] * len(self.requests))

    def test_constructor_initialiazes_a_list_of_None_for_responses_prior_to_execution(self):
        self.assertListEqual(self.batch._responses, [None] * len(self.requests))

    def test_default_failure_callback_sets_the_error_in_error_list(self):
        request_index = 3
        response = MagicMock()
        self.batch._default_failure_callback(response=response, request_index=request_index)

        self.assertEqual(
            self.batch._errors[request_index],
            FacebookBatchRequestError(request=response.request(), request_error=response.error()))

    def test_default_success_callback_sets_the_response_in_response_list(self):
        request_index = 3
        response = MagicMock()
        self.batch._default_success_callback(response=response, request_index=request_index)

        self.assertEqual(
            self.batch._responses[request_index],
            FacebookBatchResponse(request=response.request(), response=response)
        )

    def test_execute_calls_add_request_on_batch_as_many_times_as_requests_exist(self):
        self.batch.execute()
        self.batch._batch.add_request.assert_has_calls([
            call(
                request,
                success=ANY,
                failure=ANY
            ) for request_index, request in enumerate(self.requests)
        ])

    def test_execute_calls_add_execute_on_batch(self):
        self.batch.execute()
        self.batch._batch.execute.assert_called_once_with()


class TestFacebookBatchUploader(TestCase):
    def setUp(self):
        self.requests = [MagicMock()] * 49
        self.api = MagicMock()
        self.batch_uploader = FacebookBatchUploader(requests=self.requests, api=self.api)

    @patch.object(FacebookAdsApi, 'get_default_api')
    def test_constructor_will_try_to_use_default_api_if_None_is_provided(self, mock_get_default_api):
        batch_uploader = FacebookBatchUploader(requests=self.requests)
        self.assertEqual(batch_uploader.api, mock_get_default_api.return_value)

    def test_constructor_will_try_to_use_the_provided_api_if_provided(self):
        self.assertEqual(self.batch_uploader.api, self.api)

    def test_execute_raises_InvalidValueError_when_chunk_size_is_more_than_50(self):
        with self.assertRaises(InvalidValueError):
            self.batch_uploader.execute(chunk_size=51)

    def test_execute_raises_InvalidValueError_when_chunk_size_is_less_than_1(self):
        with self.assertRaises(InvalidValueError):
            self.batch_uploader.execute(chunk_size=0)

    @patch.object(FacebookBatch, 'execute')
    def test_execute_calls_execute_in_FacebookBatch(self, mock_execute):
        self.batch_uploader.execute()
        mock_execute.assert_called_once_with()

    @patch.object(FacebookBatch, 'execute')
    def test_execute_generates_as_many_batches_as_the_rounded_number_of_requests_per_chunk_size(self, mock_execute):
        chunk_size = 5
        self.batch_uploader.execute(chunk_size=chunk_size)
        self.assertEqual(len(self.batch_uploader._batches), round(len(self.requests) / chunk_size))

    @patch.object(FacebookBatchUploader, 'errors', new_callable=PropertyMock)
    @patch.object(FacebookBatch, 'execute')
    def test_execute_raises_BatchExecutionError_when_errors_exist(self, mock_execute, mock_errors):
        mock_errors.return_value = [MagicMock()] * len(self.requests)
        with self.assertRaises(BatchExecutionError):
            self.batch_uploader.execute()

    def test_requests_returns_the_same_requests_that_were_provided_to_the_constructor(self):
        self.assertListEqual(self.batch_uploader.requests, self.requests)

    def test_responses_returns_a_list_of_None_when_fetched_before_execution(self):
        self.assertListEqual(list(self.batch_uploader.responses), [])

    def test_errors_returns_an_empty_list_when_fetched_before_execution(self):
        self.assertListEqual(list(self.batch_uploader.errors), [])

    def test_items_returns_an_empty_list_when_fetched_before_execution(self):
        self.assertListEqual(list(self.batch_uploader.items), [])

    def test_errors_returns_errors_when_a_batch_has_any(self):
        facebook_batch = MagicMock()
        facebook_batch._errors = [MagicMock()] * len(self.requests)
        self.batch_uploader._batches = [facebook_batch]

        self.assertListEqual(list(self.batch_uploader.errors), facebook_batch._errors)

    def test_responses_returns_responses_when_a_batch_has_any(self):
        facebook_batch = MagicMock()
        facebook_batch._responses = [MagicMock()] * len(self.requests)
        self.batch_uploader._batches = [facebook_batch]

        self.assertListEqual(list(self.batch_uploader.responses), facebook_batch._responses)

    def test_items_returns_a_triple_with_requests_and_responses_and_errors(self):
        facebook_batch = MagicMock()
        facebook_batch._responses = [MagicMock()] * len(self.requests)
        facebook_batch._errors = [MagicMock()] * len(self.requests)
        self.batch_uploader._batches = [facebook_batch]

        self.assertListEqual(
            list(self.batch_uploader.items),
            list(zip(self.batch_uploader.requests, self.batch_uploader.responses, self.batch_uploader.errors)))

    def test_errors_returns_all_errors_when_multiple_batches_exist(self):
        facebook_batch_1 = MagicMock()
        facebook_batch_1._errors = [MagicMock()] * (len(self.requests) - 1)
        facebook_batch_2 = MagicMock()
        facebook_batch_2._errors = [MagicMock()]
        self.batch_uploader._batches = [facebook_batch_1, facebook_batch_2]

        self.assertListEqual(
            list(self.batch_uploader.errors),
            list(chain(facebook_batch_1._errors, facebook_batch_2._errors)))

    def test_responses_returns_all_responses_when_multiple_batches_exist(self):
        facebook_batch_1 = MagicMock()
        facebook_batch_1._responses = [MagicMock()] * (len(self.requests) - 1)
        facebook_batch_2 = MagicMock()
        facebook_batch_2._responses = [MagicMock()]
        self.batch_uploader._batches = [facebook_batch_1, facebook_batch_2]

        self.assertListEqual(
            list(self.batch_uploader.responses),
            list(chain(facebook_batch_1._responses, facebook_batch_2._responses)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import (
    find_packages,
    setup,
)

from wespe import (
    __author__,
    __email__,
    __version__,
)

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.read().splitlines()

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author=__author__,
    author_email=__email__,
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.6',
    ],
    keywords=('wespe adtech marketing facebook google bing yandex facebook'),
    description="Batching ad tech providers’ operations for humans",
    install_requires=requirements,
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    name='wespe',
    packages=find_packages(include=['wespe.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/x8lucas8x/wespe',
    version=__version__,
    zip_safe=False,
)
